#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>
#include <string.h>
#include "jni.h"

#ifdef MACOS
// Looks like every version of JDK from 1.7 to 16 has ../MacOS/libjli.dylib, even adoptopen
#define JLI "/../MacOS/libjli.dylib"
#endif

#ifdef MACOS_OPENJDK
#define JLI "/lib/libjli.dylib"
#endif

#ifdef LINUX
#define JLI "/jre/lib/amd64/server/libjvm.so"
#endif

int main ( int argc, char *argv[] ) {
    JavaVM         *jvm;
    JNIEnv         *jenv;
    JavaVMInitArgs  jargs;
    JavaVMOption    jopts[1];
    char            lib[1024];
    char           *className;

    className = argv[1];

    jopts[0].optionString = "-Djava.class.path=.";

    jargs.version = JNI_VERSION_1_8;
    jargs.nOptions = 1;
    jargs.options = jopts;
    jargs.ignoreUnrecognized = JNI_TRUE;

    strcpy(lib, getenv("JAVA_HOME"));
    strcat(lib, JLI);
    void * h = dlopen(lib, RTLD_LAZY);
    if ( h == NULL ) {
        printf("dlopen %s returned nil\n", lib);
        return 1;
    } else {
        printf("returned ok from dlopen %s\n", lib);
    }

    long (*crvm)() = dlsym(h, "JNI_CreateJavaVM");
    if ( crvm == NULL ) {
        printf("dlsym JNI_CreateJavaVM returned nil\n");
        return 1;
    } else {
        printf("returned ok from dlsym\n");
    }

    // This code behaves the same way whether I resolve JNI_CreateJavaVM at compile time or run time.
    long rc = crvm(&jvm, &jenv, &jargs);
    if ( rc == JNI_ERR ) {
        printf("%ld error creating vm\n", rc);
        return 1;
    } else {
        printf("returned ok from JNI_CreateJavaVM\n");
    }

    jclass cls = (*jenv)->FindClass(jenv, className);
    if ( cls == NULL ) {
        printf("FindClass(%s) failed\n", className);
        (*jenv)->ExceptionDescribe(jenv);
        return 1;
    } else {
        printf("returned ok from FindClass(%s)\n", className);
    }

    jmethodID methid = (*jenv)->GetStaticMethodID(jenv, cls, "main", "([Ljava/lang/String;)V");
    if ( methid == NULL ) {
        printf("GetStaticMethodID failed\n");
        (*jenv)->ExceptionDescribe(jenv);
        return 1;
    } else {
        printf("returned ok from GetStaticMethodID\n");
    }

    jobjectArray arr = (*jenv)->NewObjectArray(
            jenv,
            1,
            (*jenv)->FindClass(jenv, "java/lang/String"),
            (*jenv)->NewStringUTF(jenv, "Hello, world.")
            );
    (*jenv)->CallStaticVoidMethod(jenv, cls, methid, arr);
    if ( (*jenv)->ExceptionCheck(jenv) ) {
        printf("CallStaticVoidMethod failed\n");
        (*jenv)->ExceptionDescribe(jenv);
        return 1;
    } else {
        printf("returned ok from CallStaticVoidMethod\n");
    }
    return 0;
}
